open Ast
open System.IO

let prettyPrintPosUP (posup) (stream: StreamWriter) = 
(
 match posup with
  | PosUp(BaseUnary(Pred(p),Var(q))) -> stream.WriteLine(sprintf "%s(%s)" (p) (q)) 
) 

let prettyPrintTermList((tlist:TermList) , (stream:StreamWriter)) =
(
	if List.length tlist = 0 then ()
	else if (List.length tlist = 1) then ( 
		let head = List.hd tlist in (
			match head with Var(var) ->
			stream.Write (var)
		)
	)
	else (
		let head = List.hd tlist in (
			match head with Var(var) -> (
				stream.Write (var)
			) ;
			match tlist with
			|[] -> ()
			|x::tail -> (
				List.iter (
					fun x -> match x with Var(var) -> (
						stream.Write (", "^ var)
					)
				) tail;
			)
		)
	)
)

let prettyPrintBaseLiteral((bLit:BaseLiteral),(stream:StreamWriter)) =
(
	match bLit with Base(Pred(predicateName),tList) -> (
		stream.Write(predicateName);
		stream.Write("(");
		prettyPrintTermList((tList),(stream));
		stream.Write(")");
	)
)

let prettyPrintPositiveLiteral((lit:PositiveLiteral) , (stream:StreamWriter)) =
(
	match lit with PosLit(baseLiteral)
	-> prettyPrintBaseLiteral(baseLiteral , stream)
)

let prettyPrintNegativeLiteral((lit:NegativeLiteral) , (stream:StreamWriter)) =
(
	match lit with NegLit(baseLiteral)
	-> (
		stream.Write("~");
		prettyPrintBaseLiteral(baseLiteral , stream)
	)
)

let prettyPrintRule( (rule:Clause) , (stream:StreamWriter)) =
( 
	match rule with
	|Rule(positiveLiteral, posLitList, negLitList) -> (
			prettyPrintPositiveLiteral(positiveLiteral, stream);
			stream.Write(" :- ") ;
			List.iter(
				fun x -> (
				    stream.Write(" ");
					prettyPrintPositiveLiteral (x , stream)
				)
			  ) posLitList 
			;
			List.iter(
				fun x -> (
					stream.Write(" ");
					prettyPrintNegativeLiteral (x , stream);
				)
			)negLitList;
		)
	|_ -> ()
)


let prettyPrintRulewithCommas ((rule:Clause), (stream:StreamWriter)) =
(
  match rule with 
  | Rule (pLiteral, pLitList, nLitList) ->
    
     let pll = List.length(pLitList) in 
     let nll = List.length(nLitList) in
     let pArray = List.to_array(pLitList) in
     let nArray = List.to_array(nLitList) in 
      //printf "Length of positiveLitList %i\n" pll;
      //printf "Length of negativeLitList %i\n" nll;
      
      prettyPrintPositiveLiteral(pLiteral, stream);
       if ((pll <> 0) || (nll <>0)) then 
       stream.Write(" :- ");
       
       if (pll >= 2) then (
          for i = 0 to pll - 2 do
            prettyPrintPositiveLiteral (pArray.(i), stream);
            stream.Write(", ");
          done
       );
       if (pll <> 0) then      
       prettyPrintPositiveLiteral (pArray.(pll-1), stream);
       
       if (nll > 0) then stream.Write(", ");
       
       if (nll >= 2) then (
          for i = 0 to nll-2 do 
             prettyPrintNegativeLiteral (nArray.(i), stream);
             stream.Write(", ");
          done
       );
       if (nll <> 0) then 
       prettyPrintNegativeLiteral (nArray.(nll-1), stream)       
  |_ -> ()   
)

let prettyPrintProgram ( (cList:Clause List) , (stream:StreamWriter) )=
(
	List.iter(
		fun x -> ( match x with 
		    |Rule(a,b,c) -> (
			//prettyPrintRule(x,stream);
			prettyPrintRulewithCommas ((Rule(a,b,c)), (stream));
			stream.WriteLine(".")
			)
			|Query(BaseQuery(posLit)) -> (
				prettyPrintPositiveLiteral(posLit,stream);
				stream.WriteLine(" ?");
			)
			| _ -> ()
			)
	) cList
)