open Ast;
open System.IO;

type PredComparer =
class
	new() = {}
	interface System.Collections.Generic.IComparer<Predicate> with
		member self.Compare(p, q) =
		(
			match p with Pred(s) -> match q with Pred(t) -> s.CompareTo(t)
		)
	end
end

let assignUniqueIndices(edbpreds: Set<Predicate> ) =
(
	let map = ref (Map.MakeTagged(new PredComparer()).empty ) in
	(
		edbpreds.Iterate ( fun x -> 
			let n = (!map).Count in
			(
				map := map.contents.Add(x, n);
			)
		);
		!map;
	)
)

type ReachableFact = RFact of Predicate * (string List)

let renameVariables (varList:string List) =
(
	let ret = ref (Map.MakeTagged(System.StringComparer.Ordinal).empty) in
	(
		List.iter( fun x ->
			match ret.contents.TryFind(x) with
			|Some(_) -> ()
			|None -> (
				let n = ret.contents.Count in
				(
				//ultimate hack.... just to produce (char) ('a' + n) !!!!
					if (n < 26) then ret := ret.contents.Add(x,"-".Replace('-',Char.chr(Char.code('a')+n)))
					else ret := ret.contents.Add(x,"x"^string_of_int(ret.contents.Count)^"E")
				)
			)
		) varList;
	);
	!ret
)

let collectVariablesFromTermList(tlist) =
(
	let ret = ref [] in
	(
		List.iter(fun x ->
			match x with Var(s) -> ret := s::(!ret)
		) tlist;
	);
	!ret;
)
let collectVariablesFromPosLitList (posLitList) =
(
	let ret = ref [] in
	(
		List.iter(fun x ->
			match x with PosLit(Base(_,tList)) ->
			ret := List.append (!ret) (collectVariablesFromTermList(tList))
		) posLitList
	);
	!ret;
)
let collectVariablesFromNegLitList (negLitList) =
(
	let ret = ref [] in
	(
		List.iter(fun x ->
			match x with NegLit(Base(_,tList)) ->
			ret := List.append (!ret) (collectVariablesFromTermList(tList))
		) negLitList
	);
	!ret;
)
let collectVariablesFromPUPList(pupList) =
(
	let ret = ref [] in
	(
		List.iter( fun x ->
			match x with PosUp(BaseUnary(_,Var(term))) ->
			ret := term::(!ret);
		) pupList;
	);
	!ret;
)
let collectVariablesFromNUPList(nupList) =
(
	let ret = ref [] in
	(
		List.iter( fun x ->
			match x with NegUp(BaseUnary(_,Var(term))) ->
			ret := term::(!ret);
		) nupList;
	);
	!ret;
)
let rec collectVariablesFromQuery(q) =
(
	let ret = ref [] in
	(
		match q with 
		|BaseQuery(PosLit(Base(pred,tlist))) -> 
			(
				(collectVariablesFromTermList(tlist));
			)
		|ORQuery(left, right) ->
			(
				List.append (collectVariablesFromQuery(left)) (collectVariablesFromQuery(right));
			)
		|AfterQuery(left,right) ->
			(
				List.append (collectVariablesFromQuery(left)) (collectVariablesFromQuery(right));
			)
	);
)

let collectVariablesInClause (clause:Clause) =
(
	let ret = ref [] in(
		match clause with
		|Edb(PosUp(BaseUnary(pred,Var(term)))) -> (ret := term::[])
		|Rule(PosLit(Base(pred,tlist)),posLitList,negLitList) ->
			(
				ret := List.append (!ret) (collectVariablesFromTermList(tlist));
				ret := List.append (!ret) (collectVariablesFromPosLitList(posLitList));
				ret := List.append (!ret) (collectVariablesFromNegLitList(negLitList));
			)
		|NewRule(pupList, z) ->
			(
				ret := List.append (!ret) (collectVariablesFromPUPList(pupList));		
			)
		|NextRule(pupList, nupList,posLitList, negLitList) ->
			(
				ret := List.append (!ret) (collectVariablesFromPUPList(pupList));
				ret := List.append (!ret) (collectVariablesFromNUPList(nupList));
				ret := List.append (!ret) (collectVariablesFromPosLitList(posLitList));
				ret := List.append (!ret) (collectVariablesFromNegLitList(negLitList));
			)
		|Query(q) ->
			(
				ret := List.append (!ret) (collectVariablesFromQuery(q));
			)
	);
	!ret;
)
let assignVariableAName((variableName:string), (predicateIndex:int)) =
(
	variableName^string_of_int(predicateIndex)
)

let CheckAndGetFixedVariables((clause:Clause), (variableToIndex:Tagged.Map<'a,'b,'c> ), (predicateToIndex:Tagged.Map<'e,'f,'d>)) =
( (* look at rhs and find those variables bound to edb predicates,like in P(x). 
	 Construct a map, holding what value (0/1) is assigned to a particular variable
	 for a particular edb predicate. If there is a conflict, return a None.
	*)
	
	let ret = ref (Map.MakeTagged(System.StringComparer.Ordinal).empty) in
	let bad = ref false in
	(
		match clause with
		|Edb(pup) -> () //nothing on rhs.
		|Query(q) -> ()
		|NewRule(_,_) -> () //Not applicable.
		|NextRule(_,_,posLitList,negLitList) ->
			(
				List.iter ( fun x ->
					match x with PosLit(Base(Pred(predicateName),tList)) -> 
					(
						match predicateToIndex.TryFind(Pred(predicateName)) with
						|None -> ()
						|Some(z) -> (
							match List.hd tList with
							Var(varName) ->
							(
								let varIndex = variableToIndex.[varName] in
								(//varIndex = x_0E/x_1E...
									let variableNewName = assignVariableAName (varIndex,z) in
									(
										//variableNewName = x_0E0/x_0E1/x_1E0...
										//variableNewName -> 1
										match ret.contents.TryFind(variableNewName) with
										|Some(vv) ->
											(
												if ("\""^predicateName^"\""<>vv) then bad := true
											)
										|None ->
											(
												ret := ret.contents.Add(variableNewName,"\""^predicateName^"\"")
											)
									)
								)
							)
						)
					)
				) posLitList;
				
				List.iter ( fun x ->
					match x with NegLit(Base(predicate,tList)) -> 
					(
						match predicateToIndex.TryFind(predicate) with
						|None -> ()
						|Some(z) -> (
							match List.hd tList with
							Var(varName) ->
							(
								let varIndex = variableToIndex.[varName] in
								(//varIndex = x_0E/x_1E...
									let variableNewName = assignVariableAName (varIndex,z) in
									(
										//variableNewName = x_0E0/x_0E1/x_1E0...
										//variableNewName -> 0
										match ret.contents.TryFind(variableNewName) with
										|Some(vv) ->
											(
												if ("\"0\""<>vv) then bad := true
											)
										|None ->
											(
												ret := ret.contents.Add(variableNewName,"\"0\"")
											)
									)
								)
							)
						)
					)
				) negLitList
			)			
		|Rule(_, posLitList, negLitList) ->
			(
							List.iter ( fun x ->
					match x with PosLit(Base(Pred(predicateName),tList)) -> 
					(
						match predicateToIndex.TryFind(Pred(predicateName)) with
						|None -> ()
						|Some(z) -> (
							match List.hd tList with
							Var(varName) ->
							(
								let varIndex = variableToIndex.[varName] in
								(//varIndex = x_0E/x_1E...
									let variableNewName = assignVariableAName (varIndex,z) in
									(
										//variableNewName = x_0E0/x_0E1/x_1E0...
										//variableNewName -> 1
										match ret.contents.TryFind(variableNewName) with
										|Some(vv) ->
											(
												if ("\""^predicateName^"\""<>vv) then bad := true
											)
										|None ->
											(
												ret := ret.contents.Add(variableNewName,"\""^predicateName^"\"")
											)
									)
								)
							)
						)
					)
				) posLitList;
				
				List.iter ( fun x ->
					match x with NegLit(Base(predicate,tList)) -> 
					(
						match predicateToIndex.TryFind(predicate) with
						|None -> ()
						|Some(z) -> (
							match List.hd tList with
							Var(varName) ->
							(
								let varIndex = variableToIndex.[varName] in
								(//varIndex = x_0E/x_1E...
									let variableNewName = assignVariableAName (varIndex,z) in
									(
										//variableNewName = x_0E0/x_0E1/x_1E0...
										//variableNewName -> 0
										match ret.contents.TryFind(variableNewName) with
										|Some(vv) ->
											(
												if ("\"0\""<>vv) then bad := true
											)
										|None ->
											(
												ret := ret.contents.Add(variableNewName,"\"0\"")
											)
									)
								)
							)
						)
					)
				) negLitList
			)
	);
	if(!bad = true) then None
	else Some(!ret)
)

let getNewTermListFromTerm((term:Term) , (variableToIndex:Tagged.Map<string,string,'d>), (numEdbPreds:int), (fixedVariablesToValue:Tagged.Map<'a,string,'c>))=
(
	let ret = ref [] in
	(
		match term with Var(s) ->
		( 
			let renamedVariableName = variableToIndex.[s] in
			(
				for i = 0 to (numEdbPreds-1) do
					let varName = assignVariableAName(  renamedVariableName, i) in
					(
						match fixedVariablesToValue.TryFind(varName) with
						|None -> ret := List.append (!ret) (Var(varName)::[]) 
						|Some(z) -> ret := List.append (!ret) (Var(z)::[])
					)
				done
			)
		)
	);
	!ret;
)
let getNewTermListFromTermList((tlist:Term List) , (variableToIndex:Tagged.Map<string,string,'d>), (numEdbPreds:int), (fixedVariablesToValue:Tagged.Map<'a,'b,'c>))=
(
	let ret = ref[] in
	(
		List.iter(fun x ->
			ret := List.append (!ret) (getNewTermListFromTerm(x,variableToIndex,numEdbPreds,fixedVariablesToValue))
		)tlist
	);
	!ret;
)

let transformPosLiteral((posLiteral:PositiveLiteral) , (variableToIndex:Tagged.Map<string,string,'d>), 
(numEdbPreds:int), (fixedVariablesToValue:Tagged.Map<'a,'b,'c>)) =
(
	match posLiteral with PosLit(Base(predicate, tlist)) -> 
	(
		PosLit(Base(predicate, getNewTermListFromTermList(tlist, variableToIndex, numEdbPreds , fixedVariablesToValue)));
	)
)


let transformPosLitList((posLitList:PosLitList) , (variableToIndex:Tagged.Map<string,string,'d>), 
(numEdbPreds:int), (fixedVariablesToValue:Tagged.Map<'a,'b,'c>), (predicateToIndex:Tagged.Map<'p,'q,'r>)) =
(
	let ret = ref [] in
	(
		List.iter ( fun x ->
			match x with PosLit(Base(predicate,tlist)) ->
			(
				if (predicateToIndex.ContainsKey(predicate) = false)
				then 
					(
						ret := List.append (!ret) (transformPosLiteral(x,variableToIndex,numEdbPreds,fixedVariablesToValue)::[]);
					)
			)
		) posLitList
	);
	!ret;
)

let transformNegLiteral((negLiteral:NegativeLiteral) , (variableToIndex:Tagged.Map<string,string,'d>), 
(numEdbPreds:int), (fixedVariablesToValue:Tagged.Map<'a,'b,'c>)) =
(
	match negLiteral with NegLit(Base(predicate, tlist)) -> 
	(
		NegLit(Base(predicate, getNewTermListFromTermList(tlist, variableToIndex, numEdbPreds , fixedVariablesToValue)));
	)
)


let transformNegLitList((negLitList:NegLitList) , (variableToIndex:Tagged.Map<string,string,'d>), 
(numEdbPreds:int), (fixedVariablesToValue:Tagged.Map<'a,'b,'c>), (predicateToIndex:Tagged.Map<'p,'q,'r>)) =
(
	let ret = ref [] in
	(
		List.iter ( fun x ->
			match x with NegLit(Base(predicate,tlist)) ->
			(
				if (predicateToIndex.ContainsKey(predicate) = false)
				then 
					(
						ret := List.append (!ret) (transformNegLiteral(x,variableToIndex,numEdbPreds,fixedVariablesToValue)::[]);
					)
			)
		) negLitList
	);
	!ret;
)

let promoteFixedVariablesFromRightToLeft((lhsFixedVariables:Tagged.Map<'a,'b,'c> ref), (fixedVariablesToValue:Tagged.Map<'a,'b,'c>)) =
(
	fixedVariablesToValue.Iterate( fun key value ->
		if( lhsFixedVariables.contents.ContainsKey(key) = false)
			then lhsFixedVariables := lhsFixedVariables.contents.Add(key,value);
	);
)

let findActualVariableNamesOccuringOnlyInEdbPredicates((posLitList:PosLitList) , (negLitList:NegLitList) , (predicateToIndex:Tagged.Map<'a,'b,'c>) ) = 
(
	let allVariables = ref (Set.MakeTagged(System.StringComparer.Ordinal).empty) in
	let varsInIDBPreds = ref (Set.MakeTagged(System.StringComparer.Ordinal).empty) in
	(
		List.iter ( fun x ->
			match x with PosLit(Base(pred,tlist)) -> (
				List.iter ( fun y ->	match y with Var(t) ->
					allVariables := allVariables.contents.Add(t);
					if(predicateToIndex.ContainsKey(pred) = false)
					then varsInIDBPreds := varsInIDBPreds.contents.Add(t);
				) tlist;
			)
		) posLitList;
		
		List.iter ( fun x ->
			match x with NegLit(Base(pred,tlist)) -> (
				List.iter ( fun y ->	match y with Var(t) ->
					allVariables := allVariables.contents.Add(t);
					if(predicateToIndex.ContainsKey(pred) = false)
					then varsInIDBPreds := varsInIDBPreds.contents.Add(t);
				) tlist;
			)
		) negLitList;
	);
	let ret = ref (Set.MakeTagged(System.StringComparer.Ordinal).empty) in
	(
		allVariables.contents.Iterate( fun key ->
			if (varsInIDBPreds.contents.Contains(key) = false)
				then ret := ret.contents.Add(key);
		)
	);
	!ret;
)

let getReachableListForUnboundVariables ((vars:Tagged.Set<'a,'b>) , 
(variableToIndex:Tagged.Map<'a,'g,'c>) , (predicateToIndex:Tagged.Map<'d,'e,'f>)
,(fixedVariables:Tagged.Map<'m,'n,'t>)
) =
(
	let ret = ref [] in
	(
		let n = predicateToIndex.Count in
		vars.Iterate(fun x ->
			let termList = getNewTermListFromTerm(Var (x) , variableToIndex, n , fixedVariables ) in
			let rhsPred = PosLit(Base(Pred("Reachable"),termList)) in
				ret := (rhsPred :: (!ret) );	
		);
		!ret;
	)
)

let getReachableList ((posLitList:PosLitList), (negLitList:NegLitList),
(variableToIndex:Tagged.Map<'a,'g,'c>) , (predicateToIndex:Tagged.Map<'d,'e,'f>)
,(fixedVariables:Tagged.Map<'p,'q,'r>)
) =
(
	let vars = findActualVariableNamesOccuringOnlyInEdbPredicates(posLitList,negLitList,predicateToIndex) in
	let reachableList = getReachableListForUnboundVariables(vars,variableToIndex,predicateToIndex,fixedVariables) in
	reachableList;
)
let translateRestrictedFact((clause:Clause), (predicateToIndex:Tagged.Map<Predicate,int,'b>)) =
( (* Not to be worried about. This will not be translated. *)
	match clause with
	|Edb(_) -> ()
	|_ -> raise (new System.ArgumentException("operation not defined for this type") ) ;
)




let translateRestrictedNewRule((clause:Clause), (predicateToIndex:Tagged.Map<Predicate,int,'b>)) =
(
	match clause with
	|NewRule(pupList,z) -> 
		(
			let n = predicateToIndex.Count in
			let arguments = Array.create n "\"0\"" in
			(
				List.iter( fun x -> // x is a PUP.
					(
						match x with PosUp(BaseUnary(Pred(predicateName),_)) ->
						(
							let index = predicateToIndex.[Pred(predicateName)] in
							arguments.[index] <- "\""^predicateName^"\""
						)
					)
				) pupList;
			);
			let predicate = Pred("Reachable") in
			let argList = ref [] in
			(
				for i = 0 to (n-1) do
					argList := List.append (!argList) (arguments.[i]::[])
				done;
			);
			RFact(predicate , !argList)
		)
	|_ -> raise (new System.ArgumentException("Operation not defined for this type."))
)



let translateRestrictedNextRule((clause:Clause), (predicateToIndex:Tagged.Map<Predicate,int,'b>)) =
(
	(*
		Rename the variables.
		For each variable, collect the edb predicates that need to be true and/or false.
		ie; renamed_variable -> edbpredicateIndex -> (0/1)
		ie; renamed_variable_<addPredicateIndexExtension> -> (0/1)
		If for some variable x and EDB predicate P, it happens that both P(x) and ~P(x)
		then return a Nothing (clause).
		
		Produce the transformed rule, changing (renamed_var_edbpredicateindex) to 0/1 as appropriate,
		if it exists in the computed map.
	*)
	let variableCollection = collectVariablesInClause(clause) in
	let variableToIndex = renameVariables(variableCollection) in
	match CheckAndGetFixedVariables(clause,variableToIndex,predicateToIndex) with 
	Some(fixedVariablesToValue) ->
		(
		//Now check, if the LHS of the next rule doesn't have things like P(x), ~P(x).
		//Build the LHS reachable predicate, by building the argument list.
			match(clause) with
			|NextRule(pupList, nupList, posLitList,negLitList) ->
				( 
					let bad = ref false in
					let actualVarName = ref "" in
					(
						if(pupList.IsEmpty) then (
						match List.hd nupList with NegUp(BaseUnary(_,Var(aVarName)))->
							actualVarName := aVarName;
						)
						else (
						match List.hd pupList with PosUp(BaseUnary(_,Var(aVarName)))->
							actualVarName := aVarName;
						)
					);
					(
					let newVariableName = variableToIndex.[!actualVarName] in
					let lhsPredicate = Pred("Reachable") in
					let n = predicateToIndex.Count in
					let arguments = Array.create n "\"0\"" in
					let lhsFixedVariables = ref (Map.MakeTagged(System.StringComparer.Ordinal).empty) in
					(
						for i = 0 to (n-1) do
							arguments.[i] <- assignVariableAName(newVariableName , i);
						done;
						List.iter( fun x ->
							match x with PosUp(BaseUnary(Pred(predicateName),_)) ->
							(
								let predIndex = predicateToIndex.[Pred(predicateName)] in
								(
									if (false = (lhsFixedVariables.contents.ContainsKey(arguments.[predIndex])) )
										then lhsFixedVariables := (!lhsFixedVariables).Add(arguments.[predIndex], "\""^predicateName^"\"")
								)
							
							)
						) pupList;
						List.iter( fun x ->
							match x with NegUp(BaseUnary(pred,_)) ->
							(
								let predIndex = predicateToIndex.[pred] in
								(
									match lhsFixedVariables.contents.TryFind(arguments.[predIndex]) with
									|Some(z) -> 
										(
											if(z <> "\"0\"") then bad := true
										)
									|None ->
										(
											lhsFixedVariables := (!lhsFixedVariables).Add(arguments.[predIndex], "\"0\"")
										)
								)
							
							)
						) nupList;
						//Now , promote those fixed variables on rhs to lhs if they are not fixed already.
						promoteFixedVariablesFromRightToLeft(lhsFixedVariables, fixedVariablesToValue);
						
						let lhsTermList = getNewTermListFromTerm(Var(!actualVarName) , variableToIndex , n, !lhsFixedVariables) in
						let lhs = PosLit(Base(lhsPredicate,lhsTermList)) in
						(
						);
						let rhsPosLitList = transformPosLitList(posLitList,variableToIndex,n,fixedVariablesToValue,predicateToIndex) in
						let rhsNegLitList = transformNegLitList(negLitList,variableToIndex,n,fixedVariablesToValue,predicateToIndex) in
						(
						);
						let rhsPosLitList = List.append (rhsPosLitList) 
						(getReachableList(posLitList,negLitList,variableToIndex,predicateToIndex,
						fixedVariablesToValue)) in
						(
							if(!bad = true) then Edb(PosUp(BaseUnary(Pred("--Unsatisifiable Next Rule Encountered.."),Var("_"))))
							else Rule(lhs,rhsPosLitList, rhsNegLitList);
						)
					);
					);
				)
			|_ -> raise (new System.Exception("Operation not defined for this type."))
		)
	|None -> Edb(PosUp(BaseUnary(Pred("--Unsatisifiable Next Rule Encountered.."),Var("_"))))
)

let translateRestrictedRule((clause:Clause), (predicateToIndex:Tagged.Map<Predicate,int,'b>)) =
(
	let variableCollection = collectVariablesInClause(clause) in
	let variableToIndex = renameVariables(variableCollection) in
	match CheckAndGetFixedVariables(clause,variableToIndex,predicateToIndex) with 
	Some(fixedVariablesToValue) ->
		(
		//Now check, if the LHS of the next rule doesn't have things like P(x), ~P(x).
		//Build the LHS reachable predicate, by building the argument list.
			match(clause) with
			|Rule(posLiteral, posLitList,negLitList) ->
				( 
					let bad = ref false in ();
					
					(
					let lhsPredicate = Pred("Reachable") in
					let n = predicateToIndex.Count in
					let arguments = Array.create n "\"0\"" in
					let lhsFixedVariables = ref (Map.MakeTagged(System.StringComparer.Ordinal).empty) in
					(
						//Now , promote those fixed variables on rhs to lhs if they are not fixed already.
						promoteFixedVariablesFromRightToLeft(lhsFixedVariables, fixedVariablesToValue);
						
						let lhs = transformPosLiteral(posLiteral , variableToIndex, n , !lhsFixedVariables) in
						(
						);
						
						let rhsPosLitList = transformPosLitList(posLitList,variableToIndex,n,fixedVariablesToValue,predicateToIndex) in
						let rhsNegLitList = transformNegLitList(negLitList,variableToIndex,n,fixedVariablesToValue,predicateToIndex) in
						(
						);
						
						let rhsPosLitList = List.append (rhsPosLitList) 
						(getReachableList(posLitList,negLitList,variableToIndex,predicateToIndex,
						fixedVariablesToValue)) in
						(
							if(!bad = true) then Edb(PosUp(BaseUnary(Pred("--Unsatisifiable Rule Encountered.."),Var("_"))))
							else Rule(lhs,rhsPosLitList, rhsNegLitList);
						)
					);
					);
				)
			|_ -> raise (new System.Exception("Operation not defined for this type."))
		)
	|None -> Edb(PosUp(BaseUnary(Pred("--Unsatisifiable Next Rule Encountered.."),Var("_"))))
)

let translateBaseQuery(
(query:Query) , (predicateToIndex:Tagged.Map<Predicate,int,'b>) , (nextIndex:int),(nextQueryIndex:int) ) = 
(
	let clause = Query(query) in
	let variableCollection = collectVariablesInClause(clause) in
	let variableToIndex = renameVariables(variableCollection) in
	match CheckAndGetFixedVariables(clause,variableToIndex,predicateToIndex) with 
	Some(fixedVariablesToValue) ->
		(
		//Now check, if the LHS of the next rule doesn't have things like P(x), ~P(x).
		//Build the LHS reachable predicate, by building the argument list.
			match(query) with
			|BaseQuery(posLiteral) ->
				( 
					let n = predicateToIndex.Count in
					let lhsFixedVariables = ref (Map.MakeTagged(System.StringComparer.Ordinal).empty) in
					(	
						let lhs = transformPosLiteral(posLiteral , variableToIndex, n , !lhsFixedVariables) in
						(
							(Query(BaseQuery(lhs)) :: [], nextIndex, nextQueryIndex);
						)
					);
				)
			|_ -> raise (new System.Exception("Operation not defined for this type."))
		)
	|None -> (Edb(PosUp(BaseUnary(Pred("--Unsatisifiable Next Rule Encountered.."),Var("_"))))::[] , nextIndex,nextQueryIndex)
)

let getUncommonTermName((firstTermList:Set<Term>),(secondTermList:Set<Term>)) =
(
	Var ("tempXXX");
)

let rec translateRestrictedQueryRecursive(
(query:Query) , (predicateToIndex:Tagged.Map<Predicate,int,'b>) , (nextIndex:int) , (nextQueryIndex:int) ) =
(
	match query with
	|BaseQuery(q) -> ( translateBaseQuery(query, predicateToIndex, nextIndex,nextQueryIndex) )
	|ORQuery(q11,q21) -> ( 
	(* Currently translates only PosLiteral AFTER PosLiteral*)
		match (q11,q21) with
			(BaseQuery(qright),BaseQuery(qleft)) -> (
						let n = predicateToIndex.Count in ();
						let firstTermList = Set.of_list (Transformparseoutput.getVariablesList(qleft::[])) in
						let secondTermList = Set.of_list (Transformparseoutput.getVariablesList(qright::[])) in
						let commonVariables = Set.inter (firstTermList) (secondTermList)in 
						(  
							if (Set.size commonVariables = 0) then
								raise ( new System.Exception("There should be at least one shared variable in an OR query"));
							
							let queryPredicate = Pred("AutoQuery"^string_of_int(nextQueryIndex)) in
							let sharedTermsList = Set.to_list commonVariables in
							(
								let queryLiteral = PosLit(Base(queryPredicate,sharedTermsList)) in
								(
									let firstRule = Rule(queryLiteral,qleft::[],[]) in ();
									let secondRule = Rule(queryLiteral, qright::[],[]) in ();
									let thirdRule = Query(BaseQuery(queryLiteral)) in ();
									let translatedFirstRule = translateRestrictedRule (firstRule,predicateToIndex) in
									let translatedSecondRule = translateRestrictedRule(secondRule, predicateToIndex) in
									match(translateBaseQuery(BaseQuery(queryLiteral),predicateToIndex,nextIndex,nextQueryIndex+1)) 
										with(translatedThirdRule,nr1,nr2) ->
									(translatedFirstRule::(translatedSecondRule::(translatedThirdRule)) , nr1, nr2);								
								);
							)
						)
			)
			|_ -> raise (new System.ArgumentException("Cannot parse \"OR\" query with non Base-query operands. Only PosLiteral AFTER PosLiteral is supported" ));
	)
	|AfterQuery(q11,q21) -> (
	(* Currently translates only PosLiteral AFTER PosLiteral*)
		match (q11,q21) with
			(BaseQuery(qright),BaseQuery(qleft)) -> (
						let n = predicateToIndex.Count in ();
						let firstTermList = Set.of_list (Transformparseoutput.getVariablesList(qleft::[])) in
						let secondTermList = Set.of_list (Transformparseoutput.getVariablesList(qright::[])) in
						let commonVariables = Set.inter (firstTermList) (secondTermList)in 
						(  
							if (Set.size commonVariables <> 1) then
								raise ( new System.Exception("There should be exactly one shared variable in an AFTER query"));
							
							let predicateName = Pred("Autocontaminated"^string_of_int(nextIndex)) in
							let queryPredicate = Pred("AutoQuery"^string_of_int(nextQueryIndex)) in
							let sharedTerm = Set.min_elt commonVariables in
							(
								let contaminatedPredicate = PosUp(BaseUnary(predicateName,sharedTerm)) in
								let queryLiteral = PosLit(Base(queryPredicate,sharedTerm::[])) in
								(
									let firstRule = NextRule(contaminatedPredicate::[],[],qleft::[],[]) in ();
									let secondRule = Rule(queryLiteral, PosLit(Base(predicateName,sharedTerm::[]))::(qright::[]),[]) in ();
									let thirdRule = Query(BaseQuery(queryLiteral)) in ();
									let translatedFirstRule = translateRestrictedNextRule (firstRule,predicateToIndex) in
									let translatedSecondRule = translateRestrictedRule(secondRule, predicateToIndex) in
									match(translateBaseQuery(BaseQuery(queryLiteral),predicateToIndex,nextIndex+1,nextQueryIndex+1)) 
										with(translatedThirdRule,nr1,nr2) ->
									(translatedFirstRule::(translatedSecondRule::(translatedThirdRule)) , nr1, nr2);								
								);
							)
						)
			)
			|_ -> raise (new System.ArgumentException("Cannot parse \"AFTER\" query with non Base-query operands. Only PosLiteral AFTER PosLiteral is supported" ));
	)
)

module Foo_static = begin
  let counter = ref 0
  and qCounter = ref 0;
end;;

type Foo = class
  new() = { }
  member this.counter
    with get() = !Foo_static.counter;
    and set(value) = Foo_static.counter := value;
  member this.qCounter
	with get() = !Foo_static.qCounter;
	and set(value) = Foo_static.qCounter := value;
end;;

let translateRestrictedQuery((clause:Clause), (predicateToIndex:Tagged.Map<Predicate,int,'b>)) =
(
	match clause with
	|Query(q) -> (
			match ( translateRestrictedQueryRecursive(q, predicateToIndex, (new Foo()).counter,(new Foo()).qCounter ) )
				with (ret, n1, n2) -> (
					(new Foo()).counter <- n1;
					(new Foo()).qCounter <- n2;
					ret;
				)
		)
	|_ -> raise (new System.ArgumentException("operation not defined for this type"));
)

let getExtraEDBContaminatedPredicates((program:Clause List), (predicateToIndex:Tagged.Map<Predicate,int,'b>)) =
(
(* Finds the extra predicates to be added and gives them each an index.*)
	let ret = ref (Map.MakeTagged(new PredComparer()).empty) in
	let n = predicateToIndex.Count in
	let extraPredicateCount = ref 0 in
	(
		predicateToIndex.Iterate ( fun x y -> ret := ret.contents.Add(x,y) );
		List.iter(
			fun x -> match x with Query(AfterQuery(q1,q2)) ->
			(
				match (q1,q2) with (BaseQuery(qleft),BaseQuery(qright)) -> 
				(
						let firstTermList = Set.of_list (Transformparseoutput.getVariablesList(qleft::[])) in
						let secondTermList = Set.of_list (Transformparseoutput.getVariablesList(qright::[])) in
						let commonVariables = Set.inter (firstTermList) (secondTermList)in 
						(
							if (Set.size commonVariables > 1) then
								raise ( new System.Exception("The size of shared variables cannot be more than 1."));
							if (Set.size commonVariables = 1) then
							(
								let newContaminatedPredicateName = Pred("Autocontaminated"^string_of_int(!extraPredicateCount)) in
								(
									ret := ret.contents.Add(newContaminatedPredicateName,!extraPredicateCount + n);
									extraPredicateCount := 1 + !extraPredicateCount;
								)
							)
						)
				)
				|_ -> raise (new System.Exception("Cannot parse \"AFTER\" query with non Base-query operands. Only PosLiteral AFTER PosLiteral is supported"));
			)
			|_ -> ()
			
		)program;
	);
	!ret;
)

(*
	Given edb predicates, the program list.
	First assign a unique id to each of the edb predicates.
*)
let translateNegationRestrictedDatalog((program:Clause List) , edbpreds) =
(
	let ret = ref [] in
	let initialFacts = ref [] in //will have Reachable(0,1,0,1..) type of facts.
	let predicateToIndexTemp = assignUniqueIndices(edbpreds) in
	let predicateToIndex = getExtraEDBContaminatedPredicates(program,predicateToIndexTemp) in
	(
		List.iter(
			fun x ->
			(
				// x is a clause.
				match x with
				|Edb(_) -> ()
				|Query(_) -> ret := List.append (!ret) (translateRestrictedQuery(x,predicateToIndex))
				|Rule(_,_,_) -> ret := List.append (!ret)  (translateRestrictedRule(x,predicateToIndex)::[])
				|NewRule(_, _) -> initialFacts := List.append (!initialFacts) (translateRestrictedNewRule(x,predicateToIndex)::[])
				|NextRule(_,_,_,_) -> ret := List.append (!ret) (translateRestrictedNextRule(x,predicateToIndex)::[])
			)
		) program;
	);
	(!ret,!initialFacts, predicateToIndex);
)
let rec prettyPrintReachableFact((fact:ReachableFact),(stream:StreamWriter)) =
(
	match fact with RFact(Pred(predicateName),args) ->
	(
		stream.Write (predicateName);
		stream.Write (" ( ");
		let first = ref true in
		List.iter( fun x ->
			(
				if (!first=false) then stream.Write (" ,");
				first := false;
				stream.Write (x:string);
			)	
		)args;
		stream.Write (" ) .");
	)
)

let prettyPrintInitialFacts((facts:ReachableFact List),(stream:StreamWriter)) =
(
	List.iter( fun x ->
		(
			prettyPrintReachableFact (x , stream);
			stream.WriteLine ();
		)
	) facts;
)